import { __assign } from "tslib";
import { wrap, KeyTrie } from 'optimism';
import { InvariantError } from 'ts-invariant';
import { argumentsObjectFromField, canUseWeakMap, createFragmentMap, getDefaultValues, getDirectiveInfoFromField, getFragmentDefinitions, getMainDefinition, getQueryDefinition, getStoreKeyName, isEqual, isField, isInlineFragment, maybeDeepFreeze, mergeDeepArray, resultKeyNameFromField, shouldInclude, } from '../../utilities';
import { supportsResultCaching } from './entityCache';
import { fragmentMatches } from './fragments';
import { isReference, makeReference, getTypenameFromStoreObject, } from './helpers';
var StoreReader = (function () {
    function StoreReader(config) {
        var _this = this;
        var cacheKeyRoot = config && config.cacheKeyRoot || new KeyTrie(canUseWeakMap);
        this.config = __assign({ addTypename: true, cacheKeyRoot: cacheKeyRoot }, config);
        var _a = this, executeStoreQuery = _a.executeStoreQuery, executeSelectionSet = _a.executeSelectionSet, executeSubSelectedArray = _a.executeSubSelectedArray;
        this.executeStoreQuery = wrap(function (options) {
            return executeStoreQuery.call(_this, options);
        }, {
            makeCacheKey: function (_a) {
                var query = _a.query, objectOrReference = _a.objectOrReference, contextValue = _a.contextValue, variableValues = _a.variableValues;
                if (supportsResultCaching(contextValue.store)) {
                    return cacheKeyRoot.lookup(contextValue.store, query, JSON.stringify(variableValues), isReference(objectOrReference) ? objectOrReference.__ref : objectOrReference);
                }
            }
        });
        this.executeSelectionSet = wrap(function (options) {
            return executeSelectionSet.call(_this, options);
        }, {
            makeCacheKey: function (_a) {
                var selectionSet = _a.selectionSet, objectOrReference = _a.objectOrReference, execContext = _a.execContext;
                if (supportsResultCaching(execContext.contextValue.store)) {
                    return cacheKeyRoot.lookup(execContext.contextValue.store, selectionSet, JSON.stringify(execContext.variableValues), isReference(objectOrReference) ? objectOrReference.__ref : objectOrReference);
                }
            }
        });
        this.executeSubSelectedArray = wrap(function (options) {
            return executeSubSelectedArray.call(_this, options);
        }, {
            makeCacheKey: function (_a) {
                var field = _a.field, array = _a.array, execContext = _a.execContext;
                if (supportsResultCaching(execContext.contextValue.store)) {
                    return cacheKeyRoot.lookup(execContext.contextValue.store, field, array, JSON.stringify(execContext.variableValues));
                }
            }
        });
    }
    StoreReader.prototype.readQueryFromStore = function (options) {
        return this.diffQueryAgainstStore(__assign(__assign({}, options), { returnPartialData: false })).result;
    };
    StoreReader.prototype.diffQueryAgainstStore = function (_a) {
        var store = _a.store, query = _a.query, variables = _a.variables, previousResult = _a.previousResult, _b = _a.returnPartialData, returnPartialData = _b === void 0 ? true : _b, _c = _a.rootId, rootId = _c === void 0 ? 'ROOT_QUERY' : _c, config = _a.config;
        var queryDefinition = getQueryDefinition(query);
        variables = __assign(__assign({}, getDefaultValues(queryDefinition)), variables);
        var context = {
            store: store,
            dataIdFromObject: config && config.dataIdFromObject,
            cacheRedirects: (config && config.cacheRedirects) || {},
        };
        var execResult = this.executeStoreQuery({
            query: query,
            objectOrReference: rootId === 'ROOT_QUERY'
                ? makeReference('ROOT_QUERY')
                : store.get(rootId) || makeReference(rootId),
            contextValue: context,
            variableValues: variables,
        });
        var hasMissingFields = execResult.missing && execResult.missing.length > 0;
        if (hasMissingFields && !returnPartialData) {
            execResult.missing.forEach(function (info) {
                if (info.tolerable)
                    return;
                throw new InvariantError("Can't find field " + info.fieldName + " on object " + JSON.stringify(info.object, null, 2) + ".");
            });
        }
        if (previousResult) {
            if (isEqual(previousResult, execResult.result)) {
                execResult.result = previousResult;
            }
        }
        return {
            result: execResult.result,
            complete: !hasMissingFields,
        };
    };
    StoreReader.prototype.executeStoreQuery = function (_a) {
        var query = _a.query, objectOrReference = _a.objectOrReference, contextValue = _a.contextValue, variableValues = _a.variableValues;
        var mainDefinition = getMainDefinition(query);
        var fragments = getFragmentDefinitions(query);
        var fragmentMap = createFragmentMap(fragments);
        var execContext = {
            query: query,
            fragmentMap: fragmentMap,
            contextValue: contextValue,
            variableValues: variableValues,
        };
        return this.executeSelectionSet({
            selectionSet: mainDefinition.selectionSet,
            objectOrReference: objectOrReference,
            execContext: execContext,
        });
    };
    StoreReader.prototype.executeSelectionSet = function (_a) {
        var _this = this;
        var selectionSet = _a.selectionSet, objectOrReference = _a.objectOrReference, execContext = _a.execContext;
        var fragmentMap = execContext.fragmentMap, variables = execContext.variableValues;
        var finalResult = { result: null };
        var objectsToMerge = [];
        var object;
        var typename;
        if (isReference(objectOrReference)) {
            object = execContext.contextValue.store.get(objectOrReference.__ref);
            typename =
                (object && object.__typename) ||
                    (objectOrReference.__ref === 'ROOT_QUERY' && 'Query');
        }
        else {
            object = objectOrReference;
            typename = object && object.__typename;
        }
        if (this.config.addTypename) {
            var typenameFromStore = object && object.__typename;
            if (typeof typenameFromStore === 'string') {
                objectsToMerge.push({ __typename: typenameFromStore });
            }
        }
        function handleMissing(result) {
            var _a;
            if (result.missing) {
                finalResult.missing = finalResult.missing || [];
                (_a = finalResult.missing).push.apply(_a, result.missing);
            }
            return result.result;
        }
        selectionSet.selections.forEach(function (selection) {
            var _a;
            if (!shouldInclude(selection, variables)) {
                return;
            }
            if (isField(selection)) {
                var fieldResult = handleMissing(_this.executeField(object, typename, selection, execContext));
                if (typeof fieldResult !== 'undefined') {
                    objectsToMerge.push((_a = {},
                        _a[resultKeyNameFromField(selection)] = fieldResult,
                        _a));
                }
            }
            else {
                var fragment = void 0;
                if (isInlineFragment(selection)) {
                    fragment = selection;
                }
                else {
                    fragment = fragmentMap[selection.name.value];
                    if (!fragment) {
                        throw new InvariantError("No fragment named " + selection.name.value);
                    }
                }
                var match = fragmentMatches(fragment, typename, _this.config.possibleTypes);
                if (match && (object || typename === 'Query')) {
                    var fragmentExecResult = _this.executeSelectionSet({
                        selectionSet: fragment.selectionSet,
                        objectOrReference: object,
                        execContext: execContext,
                    });
                    if (match === 'heuristic' && fragmentExecResult.missing) {
                        fragmentExecResult = __assign(__assign({}, fragmentExecResult), { missing: fragmentExecResult.missing.map(function (info) {
                                return __assign(__assign({}, info), { tolerable: true });
                            }) });
                    }
                    objectsToMerge.push(handleMissing(fragmentExecResult));
                }
            }
        });
        finalResult.result = mergeDeepArray(objectsToMerge);
        if (process.env.NODE_ENV !== 'production') {
            Object.freeze(finalResult.result);
        }
        return finalResult;
    };
    StoreReader.prototype.executeField = function (object, typename, field, execContext) {
        var variables = execContext.variableValues, contextValue = execContext.contextValue;
        var fieldName = field.name.value;
        var args = argumentsObjectFromField(field, variables);
        var info = {
            resultKey: resultKeyNameFromField(field),
            directives: getDirectiveInfoFromField(field, variables),
        };
        var readStoreResult = readStoreResolver(object, typename, fieldName, args, contextValue, info);
        if (Array.isArray(readStoreResult.result)) {
            return this.combineExecResults(readStoreResult, this.executeSubSelectedArray({
                field: field,
                array: readStoreResult.result,
                execContext: execContext,
            }));
        }
        if (!field.selectionSet) {
            if (process.env.NODE_ENV !== 'production') {
                assertSelectionSetForIdValue(contextValue.store, field, readStoreResult.result);
                maybeDeepFreeze(readStoreResult);
            }
            return readStoreResult;
        }
        if (readStoreResult.result == null) {
            return readStoreResult;
        }
        return this.combineExecResults(readStoreResult, this.executeSelectionSet({
            selectionSet: field.selectionSet,
            objectOrReference: readStoreResult.result,
            execContext: execContext,
        }));
    };
    StoreReader.prototype.combineExecResults = function () {
        var execResults = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            execResults[_i] = arguments[_i];
        }
        var missing;
        execResults.forEach(function (execResult) {
            if (execResult.missing) {
                missing = missing || [];
                missing.push.apply(missing, execResult.missing);
            }
        });
        return {
            result: execResults.pop().result,
            missing: missing,
        };
    };
    StoreReader.prototype.executeSubSelectedArray = function (_a) {
        var _this = this;
        var field = _a.field, array = _a.array, execContext = _a.execContext;
        var missing;
        function handleMissing(childResult) {
            if (childResult.missing) {
                missing = missing || [];
                missing.push.apply(missing, childResult.missing);
            }
            return childResult.result;
        }
        array = array.map(function (item) {
            if (item === null) {
                return null;
            }
            if (Array.isArray(item)) {
                return handleMissing(_this.executeSubSelectedArray({
                    field: field,
                    array: item,
                    execContext: execContext,
                }));
            }
            if (field.selectionSet) {
                return handleMissing(_this.executeSelectionSet({
                    selectionSet: field.selectionSet,
                    objectOrReference: item,
                    execContext: execContext,
                }));
            }
            if (process.env.NODE_ENV !== 'production') {
                assertSelectionSetForIdValue(execContext.contextValue.store, field, item);
            }
            return item;
        });
        if (process.env.NODE_ENV !== 'production') {
            Object.freeze(array);
        }
        return { result: array, missing: missing };
    };
    return StoreReader;
}());
export { StoreReader };
function assertSelectionSetForIdValue(store, field, fieldValue) {
    if (!field.selectionSet) {
        JSON.stringify(fieldValue, function (_key, nestedValue) {
            if (isReference(nestedValue)) {
                throw new InvariantError("Missing selection set for object of type " + getTypenameFromStoreObject(store, nestedValue) + " returned for query field " + field.name.value);
            }
            return nestedValue;
        });
    }
}
function readStoreResolver(object, typename, fieldName, args, context, _a) {
    var directives = _a.directives;
    var storeKeyName = fieldName;
    if (args || directives) {
        storeKeyName = getStoreKeyName(storeKeyName, args, directives);
    }
    var fieldValue = void 0;
    if (object) {
        fieldValue = object[storeKeyName];
        if (typeof fieldValue === 'undefined' &&
            context.cacheRedirects &&
            typeof typename === 'string') {
            var type = context.cacheRedirects[typename];
            if (type) {
                var resolver = type[fieldName];
                if (resolver) {
                    fieldValue = resolver(object, args, {
                        getCacheKey: function (storeObj) {
                            var id = context.dataIdFromObject(storeObj);
                            return id && makeReference(id);
                        },
                    });
                }
            }
        }
    }
    if (typeof fieldValue === 'undefined') {
        return {
            result: fieldValue,
            missing: [{
                    object: object,
                    fieldName: storeKeyName,
                    tolerable: false,
                }],
        };
    }
    return {
        result: fieldValue,
    };
}
//# sourceMappingURL=readFromStore.js.map