import { __assign } from "tslib";
import { invariant, InvariantError } from 'ts-invariant';
import { createFragmentMap, getDefaultValues, getFragmentDefinitions, getFragmentFromSelection, getOperationDefinition, isField, resultKeyNameFromField, shouldInclude, storeKeyNameFromField, DeepMerger, getTypenameFromResult, cloneDeep, } from '../../utilities';
import { defaultNormalizedCacheFactory } from './entityCache';
import { fragmentMatches } from './fragments';
import { makeReference, isReference, getTypenameFromStoreObject, } from './helpers';
import { defaultDataIdFromObject } from './inMemoryCache';
var StoreWriter = (function () {
    function StoreWriter(config) {
        if (config === void 0) { config = {}; }
        this.config = config;
    }
    StoreWriter.prototype.writeQueryToStore = function (_a) {
        var query = _a.query, result = _a.result, _b = _a.dataId, dataId = _b === void 0 ? 'ROOT_QUERY' : _b, _c = _a.store, store = _c === void 0 ? defaultNormalizedCacheFactory() : _c, variables = _a.variables, _d = _a.dataIdFromObject, dataIdFromObject = _d === void 0 ? defaultDataIdFromObject : _d;
        var operationDefinition = getOperationDefinition(query);
        store.retain(dataId);
        return this.writeSelectionSetToStore({
            result: result,
            dataId: dataId,
            selectionSet: operationDefinition.selectionSet,
            context: {
                store: store,
                processedData: {},
                variables: __assign(__assign({}, getDefaultValues(operationDefinition)), variables),
                dataIdFromObject: dataIdFromObject,
                fragmentMap: createFragmentMap(getFragmentDefinitions(query)),
            },
        });
    };
    StoreWriter.prototype.writeSelectionSetToStore = function (_a) {
        var result = _a.result, dataId = _a.dataId, selectionSet = _a.selectionSet, context = _a.context;
        var store = context.store;
        var newFields = this.processSelectionSet({
            result: result,
            selectionSet: selectionSet,
            context: context,
        });
        store.set(dataId, mergeStoreObjects(store, store.get(dataId), newFields));
        return store;
    };
    StoreWriter.prototype.processSelectionSet = function (_a) {
        var _this = this;
        var result = _a.result, selectionSet = _a.selectionSet, context = _a.context, _b = _a.typename, typename = _b === void 0 ? getTypenameFromResult(result, selectionSet, context.fragmentMap) : _b;
        var newFields = Object.create(null);
        selectionSet.selections.forEach(function (selection) {
            if (!shouldInclude(selection, context.variables)) {
                return;
            }
            if (isField(selection)) {
                var resultFieldKey = resultKeyNameFromField(selection);
                var value = result[resultFieldKey];
                if (typeof value !== 'undefined') {
                    var storeFieldName = storeKeyNameFromField(selection, context.variables);
                    newFields[storeFieldName] = _this.processFieldValue(value, selection, context);
                }
                else if (_this.config.possibleTypes &&
                    !(selection.directives &&
                        selection.directives.some(function (_a) {
                            var name = _a.name;
                            return name && (name.value === 'defer' || name.value === 'client');
                        }))) {
                    invariant.warn("Missing field " + resultFieldKey + " in " + JSON.stringify(result, null, 2).substring(0, 100));
                }
            }
            else {
                var fragment = getFragmentFromSelection(selection, context.fragmentMap);
                var match = fragmentMatches(fragment, typename, _this.config.possibleTypes);
                if (match && (result || typename === 'Query')) {
                    Object.assign(newFields, _this.processSelectionSet({
                        result: result,
                        selectionSet: fragment.selectionSet,
                        typename: typename,
                        context: context,
                    }));
                }
            }
        });
        if (typeof typename === 'string' &&
            newFields.__typename === void 0) {
            newFields.__typename = typename;
        }
        return newFields;
    };
    StoreWriter.prototype.processFieldValue = function (value, field, context) {
        var _this = this;
        if (!field.selectionSet || value === null) {
            return process.env.NODE_ENV === 'production' ? value : cloneDeep(value);
        }
        if (Array.isArray(value)) {
            return value.map(function (item) { return _this.processFieldValue(item, field, context); });
        }
        if (value && context.dataIdFromObject) {
            var dataId = context.dataIdFromObject(value);
            if (typeof dataId === 'string') {
                if (!isDataProcessed(dataId, field, context.processedData)) {
                    this.writeSelectionSetToStore({
                        dataId: dataId,
                        result: value,
                        selectionSet: field.selectionSet,
                        context: context,
                    });
                }
                return makeReference(dataId);
            }
        }
        return this.processSelectionSet({
            result: value,
            selectionSet: field.selectionSet,
            context: context,
        });
    };
    return StoreWriter;
}());
export { StoreWriter };
function mergeStoreObjects(store, existing, incoming) {
    return new DeepMerger(function (existingObject, incomingObject, property) {
        var existing = existingObject[property];
        var incoming = incomingObject[property];
        if (existing !== incoming &&
            this.isObject(existing) &&
            this.isObject(incoming)) {
            var eType = getTypenameFromStoreObject(store, existing);
            var iType = getTypenameFromStoreObject(store, incoming);
            if (typeof eType === 'string' &&
                typeof iType === 'string' &&
                eType !== iType) {
                return incoming;
            }
            if (isReference(incoming)) {
                if (isReference(existing)) {
                    return incoming.__ref === existing.__ref ? existing : incoming;
                }
                store.set(incoming.__ref, this.merge(existing, store.get(incoming.__ref)));
                return incoming;
            }
            else if (isReference(existing)) {
                throw new InvariantError("Store error: the application attempted to write an object with no provided id but the store already contains an id of " + existing.__ref + " for this object.");
            }
            if (Array.isArray(incoming)) {
                if (!Array.isArray(existing))
                    return incoming;
                if (existing.length > incoming.length) {
                    return this.merge(existing.slice(0, incoming.length), incoming);
                }
            }
            return this.merge(existing, incoming);
        }
        return incoming;
    }).merge(existing, incoming);
}
function isDataProcessed(dataId, field, processedData) {
    var fieldSet = processedData[dataId];
    if (fieldSet) {
        if (fieldSet.has(field))
            return true;
        fieldSet.add(field);
    }
    else {
        processedData[dataId] = new Set([field]);
    }
    return false;
}
//# sourceMappingURL=writeToStore.js.map